import 'package:flutter/services.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:hive/hive.dart';
import 'package:hive_ui/core/box_update_handler.dart';
import 'package:hive_ui/core/hive_view_state.dart';
import 'package:mockito/mockito.dart';

import '../../example/lib/boxes.dart';

class BoxUpdateHandlerMock extends Mock implements BoxUpdateHandler {}

void main() {
  const String macOs = 'plugins.flutter.io/path_provider_macos';
  const String windows = 'plugins.flutter.io/path_provider';

   TestWidgetsFlutterBinding.ensureInitialized();
  const MethodChannel channel = MethodChannel(windows);
  const MethodChannel channelMac = MethodChannel(macOs);
  channelMac.setMockMethodCallHandler((MethodCall call) async {
    return ".";
  });
  channel.setMockMethodCallHandler((MethodCall methodCall) async {
    return ".";
  });
  Boxes.initHive();

  group('BoxUpdateHandler', () {
    test('addObject() should return true when object is added successfully',
            () async {
          final boxUpdateHandlerMock = BoxUpdateHandlerMock();
          final addedObject = {"id": 1, "name": "test"};
          final currentBox = await Hive.openBox('testBox');
          when(boxUpdateHandlerMock.addObject(addedObject)).thenAnswer((_) async => true);
          expect(await boxUpdateHandlerMock.addObject(addedObject), true);
          currentBox.close();
        });

    test('addObject() should return false when object is not added successfully',
            () async {
          final boxUpdateHandlerMock = BoxUpdateHandlerMock();
          final addedObject = {"id": 1, "name": "test"};
          when(boxUpdateHandlerMock.addObject(addedObject)).thenAnswer((_) async => false);
          expect(await boxUpdateHandlerMock.addObject(addedObject), false);
        });

    test(
        'addObject() should call errorCallback when exception is thrown while adding object',
            () async {
          final boxUpdateHandlerMock = BoxUpdateHandlerMock();
          final addedObject = {"id": 1, "name": "test"};
          final currentBox = await Hive.openBox('testBox');
          when(boxUpdateHandlerMock.addObject(addedObject)).thenThrow(Exception());
          expect(() => boxUpdateHandlerMock.addObject(addedObject), throwsException);
          currentBox.close();
        });
  });

  group('addObject', () {
    test('should add object to box', () async {
      final box = Hive.box('test');
      final boxHandler = BoxUpdateHandler(HiveViewState(boxesMap: {box: (json) => json}), (error) => print(error));
      final addedObject = {"id": 1, "name": "John Doe"};
      final isAdded = await boxHandler.addObject(addedObject);
      expect(isAdded, true);
      expect(box.get(1), addedObject);
    });
    test('should not add object to box and show error', () async {
      final box = Hive.box('test');
      final boxHandler = BoxUpdateHandler(HiveViewState(boxesMap: {box: (json) => json}), (error) => print(error));
      final addedObject = {"name": "John Doe"};
      final isAdded = await boxHandler.addObject(addedObject);
      expect(isAdded, false);
      expect(box.get(1), null);
    });
    test('should not add object to box and show error if fromJsonConverter is not provided', () async {
      final box = Hive.box('test');
      final boxHandler = BoxUpdateHandler(HiveViewState(boxesMap: {}), (error) => print(error));
      final addedObject = {"id": 1, "name": "John Doe"};
      final isAdded = await boxHandler.addObject(addedObject);
      expect(isAdded, false);
      expect(box.get(1), null);
    });
    test('should not add object to box if box is not opened', () async {
      final boxHandler = BoxUpdateHandler(HiveViewState(boxesMap: {}), (error) => print(error));
      final addedObject = {"id": 1, "name": "John Doe"};
      final isAdded = await boxHandler.addObject(addedObject);
      expect(isAdded, false);
    });
  });
}




